@@ -40,6 +40,9 @@
 
 #include <stdarg.h>
 
+static void
+rd_kafka_mock_cluster_destroy0 (rd_kafka_mock_cluster_t *mcluster);
+
 
 static rd_kafka_mock_broker_t *
 rd_kafka_mock_broker_find (const rd_kafka_mock_cluster_t *mcluster,
@@ -991,6 +994,21 @@ rd_kafka_mock_connection_write_out (rd_kafka_mock_connection_t *mconn) {
 }
 
 
+/**
+ * @brief Call connection_write_out() for all the broker's connections.
+ *
+ * Use to check if any responses should be sent when RTT has changed.
+ */
+static void
+rd_kafka_mock_broker_connections_write_out (rd_kafka_mock_broker_t *mrkb) {
+        rd_kafka_mock_connection_t *mconn, *tmp;
+
+        /* Need a safe loop since connections may be removed on send error */
+        TAILQ_FOREACH_SAFE(mconn, &mrkb->connections, link, tmp) {
+                rd_kafka_mock_connection_write_out(mconn);
+        }
+}
+
 
 /**
  * @brief Per-Connection IO handler
@@ -1181,6 +1199,8 @@ static int rd_kafka_mock_cluster_thread_main (void *arg) {
                                              RD_KAFKA_THREAD_BACKGROUND);
         rd_atomic32_sub(&rd_kafka_thread_cnt_curr, 1);
 
+        rd_kafka_mock_cluster_destroy0(mcluster);
+
         return 0;
 }
 
@@ -1224,14 +1244,30 @@ static void rd_kafka_mock_broker_close_all (rd_kafka_mock_broker_t *mrkb,
                 rd_kafka_mock_connection_close(mconn, reason);
 }
 
+/**
+ * @brief Destroy error stack, must be unlinked.
+ */
+static void
+rd_kafka_mock_error_stack_destroy (rd_kafka_mock_error_stack_t *errstack) {
+        if (errstack->errs)
+                rd_free(errstack->errs);
+        rd_free(errstack);
+}
+
 
 static void rd_kafka_mock_broker_destroy (rd_kafka_mock_broker_t *mrkb) {
+        rd_kafka_mock_error_stack_t *errstack;
 
         rd_kafka_mock_broker_close_all(mrkb, "Destroying broker");
 
         rd_kafka_mock_cluster_io_del(mrkb->cluster, mrkb->listen_s);
         rd_close(mrkb->listen_s);
 
+        while ((errstack = TAILQ_FIRST(&mrkb->errstacks))) {
+                TAILQ_REMOVE(&mrkb->errstacks, errstack, link);
+                rd_kafka_mock_error_stack_destroy(errstack);
+        }
+
         TAILQ_REMOVE(&mrkb->cluster->brokers, mrkb, link);
         mrkb->cluster->broker_cnt--;
 
@@ -1308,6 +1344,7 @@ rd_kafka_mock_broker_new (rd_kafka_mock_cluster_t *mcluster,
                     "%s", rd_sockaddr2str(&sin, 0));
 
         TAILQ_INIT(&mrkb->connections);
+        TAILQ_INIT(&mrkb->errstacks);
 
         TAILQ_INSERT_TAIL(&mcluster->brokers, mrkb, link);
         mcluster->broker_cnt++;
@@ -1482,17 +1519,23 @@ rd_kafka_mock_error_stack_get (rd_kafka_mock_error_stack_head_t *shead,
  * @brief Removes and returns the next request error for request type \p ApiKey.
  */
 rd_kafka_resp_err_t
-rd_kafka_mock_next_request_error (rd_kafka_mock_cluster_t *mcluster,
+rd_kafka_mock_next_request_error (rd_kafka_mock_connection_t *mconn,
                                   int16_t ApiKey) {
+        rd_kafka_mock_cluster_t *mcluster = mconn->broker->cluster;
         rd_kafka_mock_error_stack_t *errstack;
         rd_kafka_resp_err_t err;
 
         mtx_lock(&mcluster->lock);
 
-        errstack = rd_kafka_mock_error_stack_find(&mcluster->errstacks, ApiKey);
+        errstack = rd_kafka_mock_error_stack_find(&mconn->broker->errstacks,
+                                                  ApiKey);
         if (likely(!errstack)) {
-                mtx_unlock(&mcluster->lock);
-                return RD_KAFKA_RESP_ERR_NO_ERROR;
+                errstack = rd_kafka_mock_error_stack_find(&mcluster->errstacks,
+                                                          ApiKey);
+                if (likely(!errstack)) {
+                        mtx_unlock(&mcluster->lock);
+                        return RD_KAFKA_RESP_ERR_NO_ERROR;
+                }
         }
 
         err = rd_kafka_mock_error_stack_next(errstack);
@@ -1503,16 +1546,6 @@ rd_kafka_mock_next_request_error (rd_kafka_mock_cluster_t *mcluster,
 
 
 
-/**
- * @brief Destroy error stack, must be unlinked.
- */
-static void
-rd_kafka_mock_error_stack_destroy (rd_kafka_mock_error_stack_t *errstack) {
-        if (errstack->errs)
-                rd_free(errstack->errs);
-        rd_free(errstack);
-}
-
 
 void rd_kafka_mock_push_request_errors (rd_kafka_mock_cluster_t *mcluster,
                                         int16_t ApiKey, size_t cnt, ...) {
@@ -1543,6 +1576,45 @@ void rd_kafka_mock_push_request_errors (rd_kafka_mock_cluster_t *mcluster,
 }
 
 
+rd_kafka_resp_err_t
+rd_kafka_mock_broker_push_request_errors (rd_kafka_mock_cluster_t *mcluster,
+                                          int32_t broker_id,
+                                          int16_t ApiKey, size_t cnt, ...) {
+        rd_kafka_mock_broker_t *mrkb;
+        va_list ap;
+        rd_kafka_mock_error_stack_t *errstack;
+        size_t totcnt;
+
+        mtx_lock(&mcluster->lock);
+
+        if (!(mrkb = rd_kafka_mock_broker_find(mcluster, broker_id))) {
+                mtx_unlock(&mcluster->lock);
+                return RD_KAFKA_RESP_ERR__UNKNOWN_BROKER;
+        }
+
+        errstack = rd_kafka_mock_error_stack_get(&mrkb->errstacks, ApiKey);
+
+        totcnt = errstack->cnt + cnt;
+
+        if (totcnt > errstack->size) {
+                errstack->size = totcnt + 4;
+                errstack->errs = rd_realloc(errstack->errs,
+                                            errstack->size *
+                                            sizeof(*errstack->errs));
+        }
+
+        va_start(ap, cnt);
+        while (cnt-- > 0)
+                errstack->errs[errstack->cnt++] =
+                        va_arg(ap, rd_kafka_resp_err_t);
+        va_end(ap);
+
+        mtx_unlock(&mcluster->lock);
+
+        return RD_KAFKA_RESP_ERR_NO_ERROR;
+}
+
+
 void rd_kafka_mock_topic_set_error (rd_kafka_mock_cluster_t *mcluster,
                                     const char *topic,
                                     rd_kafka_resp_err_t err) {
@@ -1824,6 +1896,10 @@ rd_kafka_mock_cluster_cmd (rd_kafka_mock_cluster_t *mcluster,
                         return RD_KAFKA_RESP_ERR_BROKER_NOT_AVAILABLE;
 
                 mrkb->rtt = (rd_ts_t)rko->rko_u.mock.lo * 1000;
+
+                /* Check if there is anything to send now that the RTT
+                 * has changed or if a timer is to be started. */
+                rd_kafka_mock_broker_connections_write_out(mrkb);
                 break;
 
         case RD_KAFKA_MOCK_CMD_BROKER_SET_RACK:
@@ -1953,7 +2029,9 @@ rd_kafka_mock_cluster_destroy0 (rd_kafka_mock_cluster_t *mcluster) {
         mtx_destroy(&mcluster->lock);
 
         rd_free(mcluster->bootstraps);
-        rd_free(mcluster);
+
+        rd_close(mcluster->wakeup_fds[0]);
+        rd_close(mcluster->wakeup_fds[1]);
 }
 
 
@@ -1975,10 +2053,7 @@ void rd_kafka_mock_cluster_destroy (rd_kafka_mock_cluster_t *mcluster) {
         if (thrd_join(mcluster->thread, &res) != thrd_success)
                 rd_assert(!*"failed to join mock thread");
 
-        rd_close(mcluster->wakeup_fds[0]);
-        rd_close(mcluster->wakeup_fds[1]);
-
-        rd_kafka_mock_cluster_destroy0(mcluster);
+        rd_free(mcluster);
 }
 
 
@@ -2064,7 +2139,6 @@ rd_kafka_mock_cluster_t *rd_kafka_mock_cluster_new (rd_kafka_t *rk,
         mcluster->bootstraps = rd_malloc(bootstraps_len + 1);
         of = 0;
         TAILQ_FOREACH(mrkb, &mcluster->brokers, link) {
-                int r;
                 r = rd_snprintf(&mcluster->bootstraps[of],
                                 bootstraps_len - of,
                                 "%s%s:%d",
